@@ -1,6 +1,6 @@
 /*
   Racket
-  Copyright (c) 2004-2013 PLT Design Inc.
+  Copyright (c) 2004-2014 PLT Design Inc.
   Copyright (c) 1995-2001 Matthew Flatt
 
     This library is free software; you can redistribute it and/or
@@ -51,6 +51,22 @@ static void register_traversers(void);
 #define to_signed_hash(v) ((intptr_t)v)
 #define to_unsigned_hash(v) ((uintptr_t)v)
 
+#ifdef OBJHEAD_HAS_HASH_BITS
+/* In 3m mode, we only have 14 bits of hash code in the
+   Scheme_Object header. But the GC-level object header has some
+   leftover bits (currently 9, 11, 41, or 43, depending on the
+   platform), so use those, too. That only works for GCable
+   objects, so we use 1 of our 14 bits to indicate whether the
+   other bits are present. */
+# define GCABLE_OBJ_HASH_BIT 0x0004
+# define OBJ_HASH_USELESS_BITS 3
+#else
+# define GCABLE_OBJ_HASH_BIT 0
+# define OBJ_HASH_USELESS_BITS 2
+#endif
+#define OBJ_HASH_USEFUL_BITS (16 - OBJ_HASH_USELESS_BITS)
+#define OBJ_HASH_USEFUL_MASK ((1 << OBJ_HASH_USEFUL_BITS)-1)
+
 #ifdef MZ_PRECISE_GC
 /* keygen race conditions below are "ok", because keygen is randomness
    used to create a hashkey. Technically, a race condition allows
@@ -59,7 +75,7 @@ static void register_traversers(void);
    that only one thread at a time sets a hash code in a specific
    object, though, and watch out for a race with JIT-generated code
    running in a future and setting flags on pairs. */
-SHARED_OK static uintptr_t keygen;
+SHARED_OK static uintptr_t keygen = GCABLE_OBJ_HASH_BIT;
 
 XFORM_NONGCING static MZ_INLINE
 uintptr_t PTR_TO_LONG(Scheme_Object *o)
@@ -76,17 +92,11 @@ uintptr_t PTR_TO_LONG(Scheme_Object *o)
     uintptr_t local_keygen = keygen;
     v |= (short)local_keygen;
 #ifdef OBJHEAD_HAS_HASH_BITS
-    /* In 3m mode, we only have 14 bits of hash code in the
-       Scheme_Object header. But the GC-level object header has some
-       leftover bits (currently 9, 11, 41, or 43, depending on the
-       platform), so use those, too. That only works for GCable
-       objects, so we use 1 of our 14 bits to indicate whether the
-       other bit are present. */
     if (GC_is_allocated(o)) {
       OBJHEAD_HASH_BITS(o) = (local_keygen >> 16);
-      v |= 0x4000;
+      v |= GCABLE_OBJ_HASH_BIT;
     } else
-      v &= ~0x4000;
+      v &= ~GCABLE_OBJ_HASH_BIT;
 #endif
     if (!v) v = 0x1AD0;
 #ifdef MZ_USE_FUTURES
@@ -99,11 +109,11 @@ uintptr_t PTR_TO_LONG(Scheme_Object *o)
     } else
 #endif
       o->keyex = v;
-    keygen += 4;
+    keygen += (1 << OBJ_HASH_USELESS_BITS);
   }
 
 #ifdef OBJHEAD_HAS_HASH_BITS
-  if (v & 0x4000)
+  if (v & GCABLE_OBJ_HASH_BIT)
     bits = OBJHEAD_HASH_BITS(o);
   else
 #endif
@@ -112,7 +122,7 @@ uintptr_t PTR_TO_LONG(Scheme_Object *o)
   /* We need to drop the low two bits of `v', which
      are used for non-hashing purposes in some types. */
 
-  return (bits << 14) | ((v >> 2) & 0x3FFF);
+  return (bits << OBJ_HASH_USEFUL_BITS) | ((v >> OBJ_HASH_USELESS_BITS) & OBJ_HASH_USEFUL_MASK);
 }
 #else
 # define PTR_TO_LONG(p) ((uintptr_t)(p)>>2)
@@ -197,6 +207,15 @@ Scheme_Hash_Table *scheme_make_hash_table(int type)
   return table;
 }
 
+void scheme_clear_hash_table(Scheme_Hash_Table *ht)
+{
+  ht->size = 0;
+  ht->count = 0;
+  ht->keys = NULL;
+  ht->vals = NULL;
+  ht->mcount = 0;
+}
+
 static Scheme_Object *do_hash(Scheme_Hash_Table *table, Scheme_Object *key, int set, Scheme_Object *val)
 {
   Scheme_Object *tkey, **keys;
@@ -289,7 +308,10 @@ static Scheme_Object *do_hash(Scheme_Hash_Table *table, Scheme_Object *key, int
     Scheme_Object **oldkeys = table->keys;
     Scheme_Object **oldvals = table->vals;
 
-    size = oldsize << 1;
+    if (table->count << 1 >= table->mcount)
+      size = oldsize << 1;
+    else
+      size = oldsize;
     table->size = size;
     
     {
@@ -463,9 +485,11 @@ void scheme_hash_set_atomic(Scheme_Hash_Table *table, Scheme_Object *key, Scheme
   scheme_end_atomic_no_swap();
 }
 
-int scheme_hash_table_equal_rec(Scheme_Hash_Table *t1, Scheme_Hash_Table *t2, void *eql)
+int scheme_hash_table_equal_rec(Scheme_Hash_Table *t1, Scheme_Object *orig_t1,
+                                Scheme_Hash_Table *t2, Scheme_Object *orig_t2,
+                                void *eql)
 {
-  Scheme_Object **vals, **keys, *v;
+  Scheme_Object **vals, **keys, *val1, *val2, *key;
   int i;
 
   if ((t1->count != t2->count)
@@ -477,10 +501,21 @@ int scheme_hash_table_equal_rec(Scheme_Hash_Table *t1, Scheme_Hash_Table *t2, vo
   vals = t1->vals;
   for (i = t1->size; i--; ) {
     if (vals[i]) {
-      v = scheme_hash_get(t2, keys[i]);
-      if (!v)
+      key = keys[i];
+
+      if (!SAME_OBJ((Scheme_Object *)t1, orig_t1))
+        val1 = scheme_chaperone_hash_traversal_get(orig_t1, key, &key);
+      else
+        val1 = vals[i];
+      
+      if (!SAME_OBJ((Scheme_Object *)t2, orig_t2))
+        val2 = scheme_chaperone_hash_get(orig_t2, key);
+      else
+        val2 = scheme_hash_get(t2, key);
+
+      if (!val2)
 	return 0;
-      if (!scheme_recur_equal(vals[i], v, eql))
+      if (!scheme_recur_equal(val1, val2, eql))
 	return 0;
     }
   }
@@ -578,33 +613,62 @@ scheme_make_bucket_table (intptr_t size, int type)
   return table;
 }
 
-Scheme_Bucket_Table *scheme_clone_bucket_table(Scheme_Bucket_Table *bt)
+void scheme_clear_bucket_table(Scheme_Bucket_Table *bt)
 {
-  Scheme_Bucket_Table *table;
-  size_t asize;
+  Scheme_Bucket **ba;
 
-  table = MALLOC_ONE_TAGGED(Scheme_Bucket_Table);
-  table->so.type = scheme_bucket_table_type;
-  table->size = bt->size;
-  table->count = bt->count;
-  table->weak = bt->weak;
-  table->with_home = 0;
-  table->make_hash_indices = bt->make_hash_indices;
-  table->compare = bt->compare;
-  if (bt->mutex) {
-    Scheme_Object *sema;
-    sema = scheme_make_sema(1);
-    table->mutex = sema;
-  }
-  {
-    Scheme_Bucket **ba;
-    asize = (size_t)table->size * sizeof(Scheme_Bucket *);
-    ba = (Scheme_Bucket **)scheme_malloc(asize);
-    table->buckets = ba;
-    memcpy(ba, bt->buckets, asize);
+  bt->count = 0;
+  bt->size = 4;
+  ba = (Scheme_Bucket **)scheme_malloc(bt->size * sizeof(Scheme_Bucket **));
+  bt->buckets = ba;
+}
+
+static Scheme_Bucket *
+allocate_bucket (Scheme_Bucket_Table *table, const char *key, void *val)
+{
+  size_t bsize;
+  Scheme_Type type;
+  Scheme_Bucket *bucket;
+
+  if (table->with_home) {
+    bsize = sizeof(Scheme_Bucket_With_Home);
+    type = scheme_variable_type;
+  } else  {
+    bsize = sizeof(Scheme_Bucket);
+    type = scheme_bucket_type;
   }
 
-  return table;
+  bucket = (Scheme_Bucket *)scheme_malloc_tagged(bsize);
+
+  bucket->so.type = type;
+
+  if (type == scheme_variable_type)
+    ((Scheme_Bucket_With_Flags *)bucket)->flags = GLOB_HAS_HOME_PTR;
+
+  if (table->weak) {
+#ifdef MZ_PRECISE_GC
+    void *kb;
+    kb = GC_malloc_weak_box((void *)key, (void **)bucket, (void **)&bucket->val - (void **)bucket, 
+                            (table->weak > 1));
+    bucket->key = (char *)kb;
+#else
+    char *kb;
+    kb = (char *)MALLOC_ONE_WEAK(void *);
+    bucket->key = kb;
+    *(void **)bucket->key = (void *)key;
+    if (table->weak > 1) {
+      scheme_late_weak_reference_indirect((void **)bucket->key, (void *)key);
+      scheme_late_weak_reference_indirect((void **)&bucket->val, (void *)key);
+    } else {
+      scheme_weak_reference_indirect((void **)bucket->key, (void *)key);
+      scheme_weak_reference_indirect((void **)&bucket->val, (void *)key);
+    }
+#endif
+  } else
+    bucket->key = (char *)key;
+  bucket->val = val;
+
+  return bucket;
 }
 
 static Scheme_Bucket *
@@ -724,50 +788,10 @@ get_bucket (Scheme_Bucket_Table *table, const char *key, int add, Scheme_Bucket
     goto rehash_key;
   }
 
-  if (b) {
+  if (b)
     bucket = b;
-  } else {
-    size_t bsize;
-    Scheme_Type type;
-
-    if (table->with_home) {
-      bsize = sizeof(Scheme_Bucket_With_Home);
-      type = scheme_variable_type;
-    } else  {
-      bsize = sizeof(Scheme_Bucket);
-      type = scheme_bucket_type;
-    }
-
-    bucket = (Scheme_Bucket *)scheme_malloc_tagged(bsize);
-
-    bucket->so.type = type;
-
-    if (type == scheme_variable_type)
-      ((Scheme_Bucket_With_Flags *)bucket)->flags = GLOB_HAS_HOME_PTR;
-
-    if (table->weak) {
-#ifdef MZ_PRECISE_GC
-      void *kb;
-      kb = GC_malloc_weak_box((void *)key, (void **)bucket, (void **)&bucket->val - (void **)bucket, 
-                              (table->weak > 1));
-      bucket->key = (char *)kb;
-#else
-      char *kb;
-      kb = (char *)MALLOC_ONE_WEAK(void *);
-      bucket->key = kb;
-      *(void **)bucket->key = (void *)key;
-      if (table->weak > 1) {
-        scheme_late_weak_reference_indirect((void **)bucket->key, (void *)key);
-        scheme_late_weak_reference_indirect((void **)&bucket->val, (void *)key);
-      } else {
-        scheme_weak_reference_indirect((void **)bucket->key, (void *)key);
-        scheme_weak_reference_indirect((void **)&bucket->val, (void *)key);
-      }
-#endif
-    } else
-      bucket->key = (char *)key;
-    bucket->val = NULL;
-  }
+  else
+    bucket = allocate_bucket(table, key, NULL);
 
   table->buckets[h] = bucket;
 
@@ -835,11 +859,12 @@ scheme_change_in_table (Scheme_Bucket_Table *table, const char *key, void *naya)
     bucket->val = naya;
 }
 
-int scheme_bucket_table_equal_rec(Scheme_Bucket_Table *t1, Scheme_Bucket_Table *t2, void *eql)
+int scheme_bucket_table_equal_rec(Scheme_Bucket_Table *t1, Scheme_Object *orig_t1,
+                                  Scheme_Bucket_Table *t2, Scheme_Object *orig_t2,
+                                  void *eql)
 {
   Scheme_Bucket **buckets, *bucket;
-  void *v;
-  const char *key;
+  Scheme_Object *key, *val1, *val2;
   int i, weak, checked = 0;
 
   /* We can't compare the count values, because they're merely
@@ -857,16 +882,26 @@ int scheme_bucket_table_equal_rec(Scheme_Bucket_Table *t1, Scheme_Bucket_Table *
     bucket = buckets[i];
     if (bucket) {
       if (weak) {
-	key = (const char *)HT_EXTRACT_WEAK(bucket->key);
+	key = (Scheme_Object *)HT_EXTRACT_WEAK(bucket->key);
       } else {
-	key = bucket->key;
+	key = (Scheme_Object *)bucket->key;
       }
       if (key) {
+        if (!SAME_OBJ((Scheme_Object *)t1, orig_t1))
+          val1 = scheme_chaperone_hash_traversal_get(orig_t1, key, &key);
+        else
+          val1 = (Scheme_Object *)bucket->val;
+
 	checked++;
-	v = scheme_lookup_in_table(t2, key);
-	if (!v)
+      
+        if (!SAME_OBJ((Scheme_Object *)t2, orig_t2))
+          val2 = scheme_chaperone_hash_get(orig_t2, key);
+        else
+          val2 = (Scheme_Object *)scheme_lookup_in_table(t2, (const char *)key);
+
+	if (!val2)
 	  return 0;
-	if (!scheme_recur_equal((Scheme_Object *)bucket->val, (Scheme_Object *)v, eql))
+	if (!scheme_recur_equal(val1, val2, eql))
 	  return 0;
       }
     }
@@ -883,9 +918,9 @@ int scheme_bucket_table_equal_rec(Scheme_Bucket_Table *t1, Scheme_Bucket_Table *
     bucket = buckets[i];
     if (bucket) {
       if (weak) {
-	key = (const char *)HT_EXTRACT_WEAK(bucket->key);
+	key = (Scheme_Object *)HT_EXTRACT_WEAK(bucket->key);
       } else {
-	key = bucket->key;
+	key = (Scheme_Object *)bucket->key;
       }
       if (key) {
 	if (!checked)
@@ -903,6 +938,51 @@ int scheme_bucket_table_equal(Scheme_Bucket_Table *t1, Scheme_Bucket_Table *t2)
   return scheme_equal((Scheme_Object *)t1, (Scheme_Object *)t2);
 }
 
+Scheme_Bucket_Table *scheme_clone_bucket_table(Scheme_Bucket_Table *bt)
+{
+  Scheme_Bucket_Table *table;
+  size_t asize;
+
+  table = MALLOC_ONE_TAGGED(Scheme_Bucket_Table);
+  table->so.type = scheme_bucket_table_type;
+  table->size = bt->size;
+  table->count = bt->count;
+  table->weak = bt->weak;
+  table->with_home = 0;
+  table->make_hash_indices = bt->make_hash_indices;
+  table->compare = bt->compare;
+  if (bt->mutex) {
+    Scheme_Object *sema;
+    sema = scheme_make_sema(1);
+    table->mutex = sema;
+  }
+  {
+    Scheme_Bucket **ba, *bucket;
+    int i;
+    asize = (size_t)table->size * sizeof(Scheme_Bucket *);
+    ba = (Scheme_Bucket **)scheme_malloc(asize);
+    table->buckets = ba;
+    memcpy(ba, bt->buckets, asize);
+    /* clone individual buckets */
+    for (i = table->size; i--; ) {
+      bucket = ba[i];
+      if (bucket) {
+        if (bucket->key) {
+          if (table->weak) {
+            void *hk = (void *)HT_EXTRACT_WEAK(bucket->key);
+            if (hk)
+              bucket = allocate_bucket(table, hk, bucket->val);
+          } else
+            bucket = allocate_bucket(table, bucket->key, bucket->val);
+          ba[i] = bucket;
+        }
+      }
+    }
+  }
+
+  return table;
+}
+
 /*========================================================================*/
 /*                         precise GC hashing                             */
 /*========================================================================*/
@@ -1085,8 +1165,10 @@ XFORM_NONGCING static uintptr_t long_dbl_hash2_val(long_double d)
 static uintptr_t equal_hash_key(Scheme_Object *o, uintptr_t k, Hash_Info *hi)
 {
   Scheme_Type t;
+  Scheme_Object *orig_obj;
 
  top:
+  orig_obj = o;
   if (SCHEME_CHAPERONEP(o))
     o = ((Scheme_Chaperone *)o)->val;
   
@@ -1174,6 +1256,7 @@ static uintptr_t equal_hash_key(Scheme_Object *o, uintptr_t k, Hash_Info *hi)
   case scheme_wrap_chunk_type:
     {
       int len = SCHEME_VEC_SIZE(o), i, val;
+      Scheme_Object *elem;
 #     include "mzhashchk.inc"
 
       if (!len)
@@ -1183,11 +1266,19 @@ static uintptr_t equal_hash_key(Scheme_Object *o, uintptr_t k, Hash_Info *hi)
       --len;
       for (i = 0; i < len; i++) {
 	SCHEME_USE_FUEL(1);
-	val = equal_hash_key(SCHEME_VEC_ELS(o)[i], 0, hi);
+        if (SAME_OBJ(o, orig_obj))
+          elem = SCHEME_VEC_ELS(o)[i];
+        else
+          elem = scheme_chaperone_vector_ref(orig_obj, i);
+	val = equal_hash_key(elem, 0, hi);
 	k = (k << 5) + k + val;
       }
       
-      o = SCHEME_VEC_ELS(o)[len];
+      if (SAME_OBJ(o, orig_obj))
+        o = SCHEME_VEC_ELS(o)[len];
+      else
+        o = scheme_chaperone_vector_ref(orig_obj, len);
+
       break;
     }
   case scheme_flvector_type:
@@ -1262,7 +1353,7 @@ static uintptr_t equal_hash_key(Scheme_Object *o, uintptr_t k, Hash_Info *hi)
     {
       Scheme_Object *procs;
 
-      procs = scheme_struct_type_property_ref(scheme_equal_property, o);
+      procs = scheme_struct_type_property_ref(scheme_equal_property, orig_obj);
       if (procs) {
         Scheme_Object *a[2], *recur, *v;
         Hash_Info *hi2;
@@ -1284,7 +1375,7 @@ static uintptr_t equal_hash_key(Scheme_Object *o, uintptr_t k, Hash_Info *hi)
         }
         memcpy(hi2, hi, sizeof(Hash_Info));
 
-        a[0] = o;
+        a[0] = orig_obj;
         a[1] = recur;
         
         procs = SCHEME_VEC_ELS(procs)[2];
@@ -1316,33 +1407,41 @@ static uintptr_t equal_hash_key(Scheme_Object *o, uintptr_t k, Hash_Info *hi)
         if (!insp || scheme_inspector_sees_part(o, insp, -2)) {
           int i;
           Scheme_Structure *s1 = (Scheme_Structure *)o;
+          Scheme_Object *elem;
 	
 #         include "mzhashchk.inc"
 	
           hi->depth += 2;
 
           for (i = SCHEME_STRUCT_NUM_SLOTS(s1); i--; ) {
-            k += equal_hash_key(s1->slots[i], 0, hi);
+            if (SAME_OBJ(o, orig_obj))
+              elem = s1->slots[i];
+            else
+              elem = scheme_struct_ref(orig_obj, i);
+            k += equal_hash_key(elem, 0, hi);
             MZ_MIX(k);
           }
 	
           return k;
         } else
-          return k + (PTR_TO_LONG(o) >> 2);
+          return k + PTR_TO_LONG(o);
       }
     }
   case scheme_box_type:
     {
       SCHEME_USE_FUEL(1);
       k += 1;
-      o = SCHEME_BOX_VAL(o);
+      if (SAME_OBJ(o, orig_obj))
+        o = SCHEME_BOX_VAL(o);
+      else
+        o = scheme_unbox(orig_obj);
       hi->depth += 2;
       break;
     }
   case scheme_hash_table_type:
     {
       Scheme_Hash_Table *ht = (Scheme_Hash_Table *)o;
-      Scheme_Object **vals, **keys;
+      Scheme_Object **vals, **keys, *key, *val;
       int i;
       uintptr_t vk;
       intptr_t old_depth;
@@ -1357,9 +1456,14 @@ static uintptr_t equal_hash_key(Scheme_Object *o, uintptr_t k, Hash_Info *hi)
       vals = ht->vals;
       for (i = ht->size; i--; ) {
 	if (vals[i]) {
-          vk = equal_hash_key(keys[i], 0, hi);
+          key = keys[i];
+          if (SAME_OBJ(o, orig_obj))
+            val = vals[i];
+          else
+            val = scheme_chaperone_hash_traversal_get(orig_obj, key, &key);
+          vk = equal_hash_key(key, 0, hi);
           MZ_MIX(vk);
-	  vk += equal_hash_key(vals[i], 0, hi);
+	  vk += equal_hash_key(val, 0, hi);
           MZ_MIX(vk);
           k += vk;  /* can't mix k, because the key order shouldn't matter */
           hi->depth = old_depth; /* also needed to avoid order-sensitivity */
@@ -1384,6 +1488,8 @@ static uintptr_t equal_hash_key(Scheme_Object *o, uintptr_t k, Hash_Info *hi)
 
       for (i = scheme_hash_tree_next(ht, -1); i != -1; i = scheme_hash_tree_next(ht, i)) {
         scheme_hash_tree_index(ht, i, &ik, &iv);
+        if (!SAME_OBJ(o, orig_obj))
+          iv = scheme_chaperone_hash_traversal_get(orig_obj, ik, &ik);
         vk = equal_hash_key(ik, 0, hi);
         MZ_MIX(vk);
         vk += equal_hash_key(iv, 0, hi);
@@ -1398,7 +1504,8 @@ static uintptr_t equal_hash_key(Scheme_Object *o, uintptr_t k, Hash_Info *hi)
     {
       Scheme_Bucket_Table *ht = (Scheme_Bucket_Table *)o;
       Scheme_Bucket **buckets, *bucket;
-      const char *key;
+      const char *_key;
+      Scheme_Object *key, *val;
       int i, weak;
       uintptr_t vk;
       intptr_t old_depth;
@@ -1415,15 +1522,19 @@ static uintptr_t equal_hash_key(Scheme_Object *o, uintptr_t k, Hash_Info *hi)
       for (i = ht->size; i--; ) {
 	bucket = buckets[i];
 	if (bucket) {
-	  if (weak) {
-	    key = (const char *)HT_EXTRACT_WEAK(bucket->key);
-	  } else {
-	    key = bucket->key;
-	  }
-	  if (key) {
-	    vk = equal_hash_key((Scheme_Object *)bucket->val, 0, hi);
+	  if (weak)
+	    _key = (const char *)HT_EXTRACT_WEAK(bucket->key);
+	  else
+	    _key = bucket->key;
+	  if (_key) {
+            key = (Scheme_Object *)_key;
+            if (SAME_OBJ(o, orig_obj))
+              val = (Scheme_Object *)bucket->val;
+            else
+              val = scheme_chaperone_hash_traversal_get(orig_obj, key, &key);
+	    vk = equal_hash_key(val, 0, hi);
             MZ_MIX(vk);
-	    vk += equal_hash_key((Scheme_Object *)key, 0, hi);
+	    vk += equal_hash_key(key, 0, hi);
             MZ_MIX(vk);
             k += vk; /* can't mix k, because the key order shouldn't matter */
             hi->depth = old_depth; /* also needed to avoid order-sensitivity */
@@ -1453,7 +1564,7 @@ static uintptr_t equal_hash_key(Scheme_Object *o, uintptr_t k, Hash_Info *hi)
 	
 	return k + (MZ_OPT_HASH_KEY(&s->iso) & 0xFFFC);
       } else
-	return k + (PTR_TO_LONG(o) >> 2);
+	return k + PTR_TO_LONG(o);
     }
 # endif
   case scheme_resolved_module_path_type:
@@ -1463,6 +1574,18 @@ static uintptr_t equal_hash_key(Scheme_Object *o, uintptr_t k, Hash_Info *hi)
       o = SCHEME_PTR_VAL(o);
     }
     break;
+  case scheme_module_index_type:
+    {
+      Scheme_Modidx *midx = (Scheme_Modidx *)o;
+#     include "mzhashchk.inc"
+      hi->depth += 2;
+      k++;
+      k = (k << 3) + k;
+      k += equal_hash_key(midx->path, 0, hi);
+      o = midx->base;
+      break;
+    }
+    break;
   case scheme_place_bi_channel_type:
     {
       k += 7;
@@ -1477,7 +1600,7 @@ static uintptr_t equal_hash_key(Scheme_Object *o, uintptr_t k, Hash_Info *hi)
       if (h1)
         return h1(o, k, hi);
       else
-        return k + (PTR_TO_LONG(o) >> 2);
+        return k + PTR_TO_LONG(o);
     }
   }
 
@@ -1578,8 +1701,10 @@ static intptr_t overflow_equal_hash_key2(Scheme_Object *o, Hash_Info *hi)
 static uintptr_t equal_hash_key2(Scheme_Object *o, Hash_Info *hi)
 {
   Scheme_Type t;
+  Scheme_Object *orig_obj;
 
  top:
+  orig_obj = o;
   if (SCHEME_CHAPERONEP(o))
     o = ((Scheme_Chaperone *)o)->val;
 
@@ -1644,6 +1769,7 @@ static uintptr_t equal_hash_key2(Scheme_Object *o, Hash_Info *hi)
     {
       int len = SCHEME_VEC_SIZE(o), i;
       uintptr_t k = 0;
+      Scheme_Object *elem;
 
 #     include "mzhashchk.inc"
 
@@ -1651,7 +1777,11 @@ static uintptr_t equal_hash_key2(Scheme_Object *o, Hash_Info *hi)
 
       for (i = 0; i < len; i++) {
 	SCHEME_USE_FUEL(1);
-	k += equal_hash_key2(SCHEME_VEC_ELS(o)[i], hi);
+        if (SAME_OBJ(o, orig_obj))
+          elem = SCHEME_VEC_ELS(o)[i];
+        else
+          elem = scheme_chaperone_vector_ref(orig_obj, i);
+	k += equal_hash_key2(elem, hi);
       }
       
       return k;
@@ -1728,7 +1858,7 @@ static uintptr_t equal_hash_key2(Scheme_Object *o, Hash_Info *hi)
     {
       Scheme_Object *procs;
 
-      procs = scheme_struct_type_property_ref(scheme_equal_property, o);
+      procs = scheme_struct_type_property_ref(scheme_equal_property, orig_obj);
       if (procs) {
         Scheme_Object *a[2], *v, *recur;
         Hash_Info *hi2;
@@ -1750,7 +1880,7 @@ static uintptr_t equal_hash_key2(Scheme_Object *o, Hash_Info *hi)
         }
         memcpy(hi2, hi, sizeof(Hash_Info));
         
-        a[0] = o;
+        a[0] = orig_obj;
         a[1] = recur;
         
         procs = SCHEME_VEC_ELS(procs)[3];
@@ -1783,13 +1913,18 @@ static uintptr_t equal_hash_key2(Scheme_Object *o, Hash_Info *hi)
           int i;
           uintptr_t k = 0;
           Scheme_Structure *s1 = (Scheme_Structure *)o;
+          Scheme_Object *elem;
           
 #         include "mzhashchk.inc"
 	
           hi->depth += 2;
 
           for (i = SCHEME_STRUCT_NUM_SLOTS(s1); i--; ) {
-            k += equal_hash_key2(s1->slots[i], hi);
+            if (SAME_OBJ(o, orig_obj))
+              elem = s1->slots[i];
+            else
+              elem = scheme_struct_ref(orig_obj, i);
+            k += equal_hash_key2(elem, hi);
           }
           
           return k;
@@ -1798,13 +1933,16 @@ static uintptr_t equal_hash_key2(Scheme_Object *o, Hash_Info *hi)
       }
     }
   case scheme_box_type:
-    o = SCHEME_BOX_VAL(o);
+    if (SAME_OBJ(o, orig_obj))
+      o = SCHEME_BOX_VAL(o);
+    else
+      o = scheme_unbox(orig_obj);
     hi->depth += 2;
     goto top;
   case scheme_hash_table_type:
     {
       Scheme_Hash_Table *ht = (Scheme_Hash_Table *)o;
-      Scheme_Object **vals, **keys;
+      Scheme_Object **vals, **keys, *key, *val;
       int i;
       uintptr_t k = 0;
       intptr_t old_depth;
@@ -1818,8 +1956,13 @@ static uintptr_t equal_hash_key2(Scheme_Object *o, Hash_Info *hi)
       vals = ht->vals;
       for (i = ht->size; i--; ) {
 	if (vals[i]) {
-	  k += equal_hash_key2(keys[i], hi);
-	  k += equal_hash_key2(vals[i], hi);
+          key = keys[i];
+          if (SAME_OBJ(o, orig_obj))
+            val = vals[i];
+          else
+            val = scheme_chaperone_hash_traversal_get(orig_obj, key, &key);
+	  k += equal_hash_key2(key, hi);
+	  k += equal_hash_key2(val, hi);
           hi->depth = old_depth;
 	}
       }
@@ -1841,6 +1984,8 @@ static uintptr_t equal_hash_key2(Scheme_Object *o, Hash_Info *hi)
       
       for (i = scheme_hash_tree_next(ht, -1); i != -1; i = scheme_hash_tree_next(ht, i)) {
         scheme_hash_tree_index(ht, i, &ik, &iv);
+        if (!SAME_OBJ(o, orig_obj))
+          iv = scheme_chaperone_hash_traversal_get(orig_obj, ik, &ik);
         k += equal_hash_key2(ik, hi);
         k += equal_hash_key2(iv, hi);
         hi->depth = old_depth;
@@ -1852,7 +1997,8 @@ static uintptr_t equal_hash_key2(Scheme_Object *o, Hash_Info *hi)
     {
       Scheme_Bucket_Table *ht = (Scheme_Bucket_Table *)o;
       Scheme_Bucket **buckets, *bucket;
-      const char *key;
+      const char *_key;
+      Scheme_Object *key, *val;
       int i, weak;
       uintptr_t k = 0;
       intptr_t old_depth;
@@ -1868,14 +2014,18 @@ static uintptr_t equal_hash_key2(Scheme_Object *o, Hash_Info *hi)
       for (i = ht->size; i--; ) {
 	bucket = buckets[i];
 	if (bucket) {
-	  if (weak) {
-	    key = (const char *)HT_EXTRACT_WEAK(bucket->key);
-	  } else {
-	    key = bucket->key;
-	  }
-	  if (key) {
-	    k += equal_hash_key2((Scheme_Object *)bucket->val, hi);
-	    k += equal_hash_key2((Scheme_Object *)key, hi);
+	  if (weak)
+	    _key = (const char *)HT_EXTRACT_WEAK(bucket->key);
+	  else
+	    _key = bucket->key;
+	  if (_key) {
+            key = (Scheme_Object *)_key;
+            if (SAME_OBJ(o, orig_obj))
+              val = (Scheme_Object *)bucket->val;
+            else
+              val = scheme_chaperone_hash_traversal_get(orig_obj, key, &key);
+	    k += equal_hash_key2(val, hi);
+	    k += equal_hash_key2(key, hi);
             hi->depth = old_depth;
 	  }
 	}
@@ -1887,6 +2037,16 @@ static uintptr_t equal_hash_key2(Scheme_Object *o, Hash_Info *hi)
     /* Needed for interning */
     o = SCHEME_PTR_VAL(o);
     goto top;
+  case scheme_module_index_type:
+    {
+      Scheme_Modidx *midx = (Scheme_Modidx *)o;
+      uintptr_t v1, v2;
+#     include "mzhashchk.inc"
+      hi->depth += 2;
+      v1 = equal_hash_key2(midx->path, hi);
+      v2 = equal_hash_key2(midx->base, hi);
+      return v1 + v2;
+    }
   case scheme_place_bi_channel_type:
     /* a bi channel has sendch and recvch, but
        sends are the same iff recvs are the same: */
@@ -1930,18 +2090,36 @@ typedef struct AVLNode {
 #if 0
 # define AVL_ASSERT(p) if (p) { } else { scheme_signal_error("hash-tree assert failure %d", __LINE__); }
 # define AVL_ASSERT_ONLY(x) x
+# define AVL_CHECK_FORMS 1
 #else
 # define AVL_ASSERT(p) /* empty */
 # define AVL_ASSERT_ONLY(x) /* empty */
 #endif
 
+XFORM_NONGCING static int get_height(AVLNode* t)
+{
+  if (t == NULL)
+    return 0;
+  else
+    return t->height;
+}
+
+XFORM_NONGCING static void fix_height(AVLNode* t)
+{
+  int h;
+  h = get_height(t->left);
+  if (get_height(t->right) > h)
+    h = get_height(t->right);
+  t->height = h + 1;
+}
+
 static AVLNode *make_avl(AVLNode *left,
                          uintptr_t code, Scheme_Object *key, Scheme_Object *val,
                          AVLNode *right)
 {
   AVLNode *avl;
 
-  avl = MALLOC_ONE_TAGGED(AVLNode);
+  avl = scheme_malloc_small_dirty_tagged(sizeof(AVLNode));
   SET_REQUIRED_TAG(avl->type = scheme_rt_avl_node);
   avl->code = code;
   avl->key = key;
@@ -1949,6 +2127,8 @@ static AVLNode *make_avl(AVLNode *left,
   avl->left = left;
   avl->right = right;
 
+  fix_height(avl);
+
   return avl;
 }
 
@@ -1960,28 +2140,11 @@ static AVLNode *avl_clone(AVLNode *avl)
   return naya;
 }
 
-XFORM_NONGCING static int get_height(AVLNode* t)
-{
-  if (t == NULL)
-    return 0;
-  else
-    return t->height;
-}
-
 XFORM_NONGCING static int get_balance(AVLNode* t)
 {
   return get_height(t->left) - get_height(t->right);
 }
 
-XFORM_NONGCING static void fix_height(AVLNode* t)
-{
-  int h;
-  h = get_height(t->left);
-  if (get_height(t->right) > h)
-    h = get_height(t->right);
-  t->height = h + 1;
-}
-
 XFORM_NONGCING static AVLNode *avl_find(uintptr_t code, AVLNode *s)
 {
   while (1) {
@@ -1997,11 +2160,28 @@ XFORM_NONGCING static AVLNode *avl_find(uintptr_t code, AVLNode *s)
   }
 }
 
+#ifdef AVL_CHECK_FORMS
 static AVLNode *AVL_CHK(AVLNode *avl, uintptr_t code)
 {
   AVL_ASSERT(avl_find(code, avl));
   return avl;
 }
+static void AVL_CHK_FORM(AVLNode *avl)
+{
+  if (avl) {
+    int h1, h2;
+    h1 = get_height(avl->left);
+    h2 = get_height(avl->right);
+    if (h2 > h1) h1 = h2;
+    AVL_ASSERT(avl->height == (h1 + 1));
+    AVL_CHK_FORM(avl->left);
+    AVL_CHK_FORM(avl->right);
+  }
+}
+#else
+# define AVL_CHK(avl, code) avl
+XFORM_NONGCING static void AVL_CHK_FORM(AVLNode *avl)  { }
+#endif
   
 AVLNode* check_rotate_right(AVLNode* t)
 {
@@ -2243,6 +2423,8 @@ static void *hash_tree_set(Scheme_Hash_Tree *tree, Scheme_Object *key, Scheme_Ob
   AVLNode *added;
   int delta;
 
+  AVL_CHK_FORM(root);
+
   if (!val) {
     /* Removing ... */
     added = avl_find(h, root);
@@ -2261,6 +2443,8 @@ static void *hash_tree_set(Scheme_Hash_Tree *tree, Scheme_Object *key, Scheme_Ob
         if (tree) {
           tree2 = MALLOC_ONE_TAGGED(Scheme_Hash_Tree);
           memcpy(tree2, tree, sizeof(Scheme_Hash_Tree));
+
+          AVL_CHK_FORM(root);
           
           tree2->root = root;
           --tree2->count;
@@ -2355,6 +2539,8 @@ static void *hash_tree_set(Scheme_Hash_Tree *tree, Scheme_Object *key, Scheme_Ob
     delta = 1;
   }
 
+  AVL_CHK_FORM(root);
+
   if (tree) {
     tree2 = MALLOC_ONE_TAGGED(Scheme_Hash_Tree);
     memcpy(tree2, tree, sizeof(Scheme_Hash_Tree));
@@ -2542,7 +2728,9 @@ int scheme_hash_tree_index(Scheme_Hash_Tree *tree, mzlonglong pos, Scheme_Object
   return path_find(tree->root, pos, _key, _val);
 }
 
-int scheme_hash_tree_equal_rec(Scheme_Hash_Tree *t1, Scheme_Hash_Tree *t2, void *eql)
+int scheme_hash_tree_equal_rec(Scheme_Hash_Tree *t1, Scheme_Object *orig_t1,
+                               Scheme_Hash_Tree *t2, Scheme_Object *orig_t2,
+                               void *eql)
 {
   Scheme_Object *k, *v, *v2;
   int i;
@@ -2553,7 +2741,15 @@ int scheme_hash_tree_equal_rec(Scheme_Hash_Tree *t1, Scheme_Hash_Tree *t2, void
     
   for (i = scheme_hash_tree_next(t1, -1); i != -1; i = scheme_hash_tree_next(t1, i)) {
     scheme_hash_tree_index(t1, i, &k, &v);
-    v2 = scheme_hash_tree_get(t2, k);
+
+    if (!SAME_OBJ((Scheme_Object *)t1, orig_t1))
+      v = scheme_chaperone_hash_traversal_get(orig_t1, k, &k);
+      
+    if (!SAME_OBJ((Scheme_Object *)t2, orig_t2))
+      v2 = scheme_chaperone_hash_get(orig_t2, k);
+    else
+      v2 = scheme_hash_tree_get(t2, k);
+
     if (!v2)
       return 0;
     if (!scheme_recur_equal(v, v2, eql))
