@@ -31,7 +31,13 @@
 
 #ifdef __GNUC__
 #pragma GCC diagnostic ignored "-Waddress"
-#pragma GCC diagnostic ignored "-Wpointer-to-int-cast"
+#ifdef __clang__
+#  ifdef MZ_PRECISE_GC
+#   pragma clang diagnostic ignored "-Wtautological-compare"
+#   pragma clang diagnostic ignored "-Wself-assign"
+#   pragma clang diagnostic ignored "-Wconstant-logical-operand"
+#  endif
+# endif
 #endif
 
 /* Separate JIT_PRECISE_GC lets us test some 3m support in non-3m mode: */
@@ -67,9 +73,6 @@ END_XFORM_ARITH;
 #endif
 
 #ifdef MZ_USE_JIT_PPC
-# ifndef DEFINE_LIGHTNING_FUNCS
-#  define SUPPRESS_LIGHTNING_FUNCS
-# endif
 # define DEFINE_LIGHTNING_FUNCS_STATIC /* empty */
 # define jit_notify_freed_code scheme_jit_notify_freed_code
 # define jit_flush_code scheme_jit_flush_code
@@ -77,8 +80,13 @@ END_XFORM_ARITH;
 # define _jit_epilog scheme_jit_epilog
 #endif
 
+#ifndef DEFINE_LIGHTNING_FUNCS
+# define SUPPRESS_LIGHTNING_FUNCS
+#endif
+
 #include "lightning/lightning.h"
 #define _jit (jitter->js)
+#define _jitp (&_jit)
 
 #ifdef MZ_USE_JIT_X86_64
 # define JIT_LOG_WORD_SIZE 3
@@ -128,20 +136,16 @@ END_XFORM_ARITH;
 #if defined(MZ_USE_JIT_I386) && !defined(MZ_USE_JIT_X86_64)
 # define USE_TINY_JUMPS
 #endif
-
-#if defined(MZ_PRECISE_GC) && defined(MZ_USE_JIT_I386)
-# define USE_FLONUM_UNBOXING
+#if defined(MZ_USE_JIT_ARM)
+/* For ARM, long jumps are needed for jumps longer than 2^23: */
+# define NEED_LONG_JUMPS
+# define LONG_JUMPS_DEFAULT(x) 1
 #endif
 
 #ifdef MZ_USE_FUTURES
 # define MZ_USE_LWC
 #endif
 
-#define JIT_NOT_RET JIT_R1
-#if JIT_NOT_RET == JIT_RET
-Fix me! See use.
-#endif
-
 #ifdef MZ_USE_SINGLE_FLOATS
 # define SCHEME_FLOAT_TYPE scheme_float_type
 #else
@@ -162,6 +166,39 @@ Fix me! See use.
 # endif
 #endif
 
+#if defined(CAN_INLINE_ALLOC)
+# if defined(MZ_USE_JIT_I386)
+#  define USE_FLONUM_UNBOXING
+# endif
+# if defined(MZ_USE_JIT_ARM) && defined(__ARM_PCS_VFP)
+#  define USE_FLONUM_UNBOXING
+# endif
+#endif
+
+#if defined(__GNUC__)
+# define USED_ONLY_SOMETIMES __attribute__((unused))
+#else
+# define USED_ONLY_SOMETIMES /* empty */
+#endif
+
+#if !defined(MZ_USE_FUTURES)
+# define USED_ONLY_FOR_FUTURES USED_ONLY_SOMETIMES
+#else
+# define USED_ONLY_FOR_FUTURES /* empty */
+#endif
+
+#if !defined(USE_FLONUM_UNBOXING)
+# define USED_ONLY_IF_FLONUM_UNBOXING USED_ONLY_SOMETIMES
+#else
+# define USED_ONLY_IF_FLONUM_UNBOXING /* empty */
+#endif
+
+#if !defined(MZ_LONG_DOUBLE)
+# define USED_ONLY_IF_LONG_DOUBLE USED_ONLY_SOMETIMES
+#else
+# define USED_ONLY_IF_LONG_DOUBLE /* empty */
+#endif
+
 #include "jitfpu.h"
 
 #if 0
@@ -177,12 +214,12 @@ static void assert_failure(int where) { printf("JIT assert failed %d\n", where);
 extern int jit_sizes[NUM_CATEGORIES];
 extern int jit_counts[NUM_CATEGORIES];
 extern int jit_code_size;
-# define START_JIT_DATA() void *__pos = jit_get_ip().ptr; uintptr_t __total = 0
+# define START_JIT_DATA() void *__pos = jit_get_ip(); uintptr_t __total = 0
 # define END_JIT_DATA(where) if (jitter->retain_start) { \
-                              jit_sizes[where] += __total + ((uintptr_t)jit_get_ip().ptr - (uintptr_t)__pos); \
+                              jit_sizes[where] += __total + ((uintptr_t)jit_get_ip() - (uintptr_t)__pos); \
                               jit_counts[where]++; }
-# define PAUSE_JIT_DATA() __total += ((uintptr_t)jit_get_ip().ptr - (uintptr_t)__pos)
-# define RESUME_JIT_DATA() __pos = jit_get_ip().ptr
+# define PAUSE_JIT_DATA() __total += ((uintptr_t)jit_get_ip() - (uintptr_t)__pos)
+# define RESUME_JIT_DATA() __pos = jit_get_ip()
 # define RECORD_CODE_SIZE(s) jit_code_size += s
 #else
 # define START_JIT_DATA() /* empty */
@@ -329,6 +366,9 @@ struct scheme_jit_common_record {
   void *retry_alloc_code_keep_extfpr1;
 # endif
 #endif
+  void *make_rest_list_code, *make_rest_list_clear_code;
+  void *call_check_not_defined_code, *call_check_assign_not_defined_code;
+  void *force_value_same_mark_code;
 
   Continuation_Apply_Indirect continuation_apply_indirect_code;
 #ifdef MZ_USE_LWC
@@ -555,8 +595,8 @@ void *scheme_jit_get_threadlocal_table();
 #  define tl_scheme_jit_save_extfp (&scheme_jit_save_extfp)
 #  define tl_scheme_jit_save_extfp2 (&scheme_jit_save_extfp2)
 # endif
-# define tl_scheme_fuel_counter (&scheme_fuel_counter)
-# define tl_scheme_jit_stack_boundary (&scheme_jit_stack_boundary)
+# define tl_scheme_fuel_counter ((void *)&scheme_fuel_counter)
+# define tl_scheme_jit_stack_boundary ((void *)&scheme_jit_stack_boundary)
 #endif
 
 /*========================================================================*/
@@ -709,6 +749,8 @@ int check_location;
       is one word past alignment. Push 1 to realign (but
       mz_push_locals() pushes 3, because we need at least
       two locals). 
+    - On ARM, the stack should be 8-byte aligned, and
+      jit_prolog() leaves the stack in an aligned state.
 */
 
 /*    LOCAL1 is used to save the value current_cont_mark_stack,
@@ -740,7 +782,7 @@ int check_location;
 
    * On some platforms, a lightweight function created with
      mz_prolog() and mz_epilog() uses LOCAL2 to save the return
-     address. On those platforms, though, LOCAL3 is dufferent from
+     address. On those platforms, though, LOCAL3 is different from
      LOCAL2. So, LOCAL3 can always be used for temporary storage in
      such functions (assuming that they're called from a function that
      pushes locals, and that nothing else is using LOCAL2).
@@ -773,7 +815,8 @@ int check_location;
 #define mz_set_local_p(x, l) mz_set_local_p_x(x, l, JIT_FP)
 #define mz_get_local_p(x, l) mz_get_local_p_x(x, l, JIT_FP)
 
-#ifdef MZ_USE_JIT_PPC
+/* --- PPC --- */
+#if defined(MZ_USE_JIT_PPC)
 /* JIT_LOCAL1, JIT_LOCAL2, and JIT_LOCAL3 are offsets in the stack frame. */
 # define JIT_LOCAL1 56
 # define JIT_LOCAL2 60
@@ -821,7 +864,41 @@ void scheme_jit_prolog_again(mz_jit_state *jitter, int n, int ret_addr_reg)
 }
 # endif
 # define _jit_prolog_again scheme_jit_prolog_again
-#else
+#endif
+
+/* --- ARM --- */
+#ifdef MZ_USE_JIT_ARM
+# define JIT_LOCAL1 JIT_FRAME_EXTRA_SPACE_OFFSET
+# define JIT_LOCAL2 (JIT_FRAME_EXTRA_SPACE_OFFSET+4)
+# define JIT_LOCAL3 (JIT_FRAME_EXTRA_SPACE_OFFSET+8)
+# define JIT_LOCAL4 (JIT_FRAME_EXTRA_SPACE_OFFSET+12)
+# define JIT_FRAME_FLOSTACK_OFFSET JIT_FRAME_EXTRA_SPACE_OFFSET
+# define mz_set_local_p_x(x, l, FP) jit_stxi_p(l, FP, x)
+# define mz_get_local_p_x(x, l, FP) jit_ldxi_p(x, FP, l)
+# define mz_patch_branch_at(a, v) jit_patch_at(a, v)
+# define mz_patch_ucbranch_at(a, v) jit_patch_at(a, v)
+# define mz_prolog(x) (mz_set_local_p(JIT_LR, JIT_LOCAL2))
+# define mz_epilog(x) (mz_get_local_p(JIT_LR, JIT_LOCAL2), jit_jmpr(JIT_LR))
+# define mz_epilog_without_jmp() /* empty */
+# define jit_shuffle_saved_regs() /* empty */
+# define jit_unshuffle_saved_regs() /* empty */
+# define mz_push_locals() /* empty */
+# define mz_pop_locals() /* empty */
+# define jit_base_prolog() jit_prolog(0)
+# ifdef SUPPRESS_LIGHTNING_FUNCS
+void scheme_jit_prolog_again(mz_jit_state *jitter, int n, int ret_addr_reg);
+# else
+void scheme_jit_prolog_again(mz_jit_state *jitter, int n, int ret_addr_reg)
+{
+  jit_movr_p(JIT_LR, ret_addr_reg);
+  arm_prolog(_jitp, n);
+}
+# endif
+# define _jit_prolog_again scheme_jit_prolog_again
+#endif
+
+/* --- x86[_64] --- */
+#if defined(JIT_X86_64) || defined(JIT_X86_PLAIN)
 /* From frame pointer, -1 is saved frame pointer, -2 is saved ESI/R12,
    and -3 is saved EDI/R13. On entry to a procedure, prolog pushes 4
    since the call (which also pushed), so if the stack was 16-bytes
@@ -934,8 +1011,8 @@ static jit_insn *fp_tmpr;
 # define mz_st_fppop_x(i, r, FP, extfl) (check_fp_depth(i, FP), (void)jit_FPSEL_stxi_xd_fppop(extfl, (JIT_FRAME_FLOSTACK_OFFSET - (i)), FP, r))
 # define mz_st_fppop(i, r, extfl) mz_st_fppop_x(i, r, JIT_FP, extfl) 
 
-#define mz_patch_branch(a) mz_patch_branch_at(a, (_jit.x.pc))
-#define mz_patch_ucbranch(a) mz_patch_ucbranch_at(a, (_jit.x.pc))
+#define mz_patch_branch(a) mz_patch_branch_at(a, jit_get_ip())
+#define mz_patch_ucbranch(a) mz_patch_ucbranch_at(a, jit_get_ip())
 
 #ifdef NEED_LONG_JUMPS
 # define __START_SHORT_JUMPS__(cond) if (cond) { _jitl.long_jumps = 0; }
@@ -969,15 +1046,6 @@ static jit_insn *fp_tmpr;
 # define __END_TINY_JUMPS_IF_COMPACT__(cond) __END_TINY_JUMPS__(cond)
 #endif
 
-/* mz_b..i_p supports 64-bit constants on x86_64: */
-#ifdef MZ_USE_JIT_X86_64
-# define mz_beqi_p(a, v, i) ((void)jit_patchable_movi_p(JIT_REXTMP, i), jit_beqr_p(a, v, JIT_REXTMP))
-# define mz_bnei_p(a, v, i) ((void)jit_patchable_movi_p(JIT_REXTMP, i), jit_bner_p(a, v, JIT_REXTMP))
-#else
-# define mz_beqi_p(a, v, i) jit_beqi_p(a, v, i)
-# define mz_bnei_p(a, v, i) jit_bnei_p(a, v, i)
-#endif
-
 #ifdef jit_leai_l
 # define jit_fixnum_l(JIT_Rdest, JIT_Rsrc) jit_leai_l(JIT_Rdest, JIT_Rsrc, 1, 1)
 #else
@@ -1042,18 +1110,18 @@ static jit_insn *fp_tmpr;
 
 #ifdef MZ_USE_LWC
 # ifdef JIT_RUNSTACK_BASE
-#  define SAVE_RS_BASE_REG() jit_stxi_p((int)&((Scheme_Current_LWC *)0x0)->runstack_base_end, JIT_R0, JIT_RUNSTACK_BASE)
+#  define SAVE_RS_BASE_REG() jit_stxi_p((intptr_t)&((Scheme_Current_LWC *)0x0)->runstack_base_end, JIT_R0, JIT_RUNSTACK_BASE)
 # else
 #  define SAVE_RS_BASE_REG() (void)0
 # endif
 # define adjust_lwc_return_address(pc) ((jit_insn *)((char *)(pc) - jit_return_pop_insn_len()))
 # define mz_finish_lwe(d, refr) (mz_tl_ldi_p(JIT_R0, tl_scheme_current_lwc), \
-                                 jit_stxi_p((int)&((Scheme_Current_LWC *)0x0)->frame_end, JIT_R0, JIT_FP), \
-                                 jit_stxi_p((int)&((Scheme_Current_LWC *)0x0)->stack_end, JIT_R0, JIT_SP), \
-                                 jit_stxi_p((int)&((Scheme_Current_LWC *)0x0)->saved_v1, JIT_R0, JIT_V1), \
+                                 jit_stxi_p((intptr_t)&((Scheme_Current_LWC *)0x0)->frame_end, JIT_R0, JIT_FP), \
+                                 jit_stxi_p((intptr_t)&((Scheme_Current_LWC *)0x0)->stack_end, JIT_R0, JIT_SP), \
+                                 jit_stxi_p((intptr_t)&((Scheme_Current_LWC *)0x0)->saved_v1, JIT_R0, JIT_V1), \
                                  SAVE_RS_BASE_REG(),                    \
                                  refr = jit_patchable_movi_p(JIT_R1, jit_forward()), \
-                                 jit_stxi_p((int)&((Scheme_Current_LWC *)0x0)->original_dest, JIT_R0, JIT_R1), \
+                                 jit_stxi_p((intptr_t)&((Scheme_Current_LWC *)0x0)->original_dest, JIT_R0, JIT_R1), \
                                  mz_finish(d),                          \
                                  jit_patch_movi(refr, adjust_lwc_return_address(_jit.x.pc)))
 #else
@@ -1135,7 +1203,8 @@ static void emit_indentation(mz_jit_state *jitter)
 #define JIT_FPU_FPR_1(r) JIT_FPU_FPR1
 #endif
 
-#if defined(MZ_USE_JIT_I386)
+#if defined(MZ_USE_JIT_I386) && (!defined(JIT_X86_64) || !defined(JIT_X86_SSE))
+/* This is better than lightning's x87 or 32-bit SSE jit_movi_d[_fppush](): */
 # define mz_movi_d_fppush(rd,immd,tmp)    { GC_CAN_IGNORE void *addr; \
                                             addr = scheme_mz_retain_double(jitter, immd); \
                                             (void)jit_patchable_movi_p(tmp, addr);        \
@@ -1233,14 +1302,14 @@ static void emit_indentation(mz_jit_state *jitter)
 # define JIT_BUFFER_PAD_SIZE 100
 #endif
 
-#define PAST_LIMIT() ((uintptr_t)jit_get_ip().ptr > (uintptr_t)jitter->limit)
+#define PAST_LIMIT() ((uintptr_t)jit_get_raw_ip() > (uintptr_t)jitter->limit)
 #define CHECK_LIMIT() if (PAST_LIMIT()) return past_limit(jitter, __FILE__, __LINE__);
 #if 1
 # define past_limit(j, f, l) 0
 #else
 static int past_limit(mz_jit_state *jitter, const char *file, int line)
 {
-  if (((uintptr_t)jit_get_ip().ptr > (uintptr_t)jitter->limit + JIT_BUFFER_PAD_SIZE)
+  if (((uintptr_t)jit_get_raw_ip() > (uintptr_t)jitter->limit + JIT_BUFFER_PAD_SIZE)
       || (jitter->retain_start)) {
     printf("way past %s %d\n", file, line); abort();
   }
@@ -1248,6 +1317,16 @@ static int past_limit(mz_jit_state *jitter, const char *file, int line)
 }
 #endif
 
+/* Use CHECK_NESTED_GENERATE() after a nested call to scheme_generate_one()
+   or after getting a shared code pointer that may be generated by another
+   place: */
+#ifdef SET_DEFAULT_LONG_JUMPS
+extern int scheme_check_long_mode(int long_mode);
+# define CHECK_NESTED_GENERATE() if (scheme_check_long_mode(_jitl.long_jumps_default)) return 0;
+#else
+# define CHECK_NESTED_GENERATE() /* empty */
+#endif
+
 void *scheme_generate_one(mz_jit_state *old_jitter, 
 			  Generate_Proc generate,
 			  void *data,
@@ -1388,7 +1467,8 @@ int scheme_generate_app(Scheme_App_Rec *app, Scheme_Object **alt_rands, int num_
 			mz_jit_state *jitter, int is_tail, int multi_ok, int ignored_result,
                         int no_call);
 int scheme_generate_tail_call(mz_jit_state *jitter, int num_rands, int direct_native, int need_set_rs, 
-                              int is_inline, Scheme_Native_Closure *direct_to_code, jit_direct_arg *direct_arg);
+                              int is_inline, Scheme_Native_Closure *direct_to_code, jit_direct_arg *direct_arg,
+                              Scheme_Closure_Data *direct_data);
 int scheme_generate_non_tail_call(mz_jit_state *jitter, int num_rands, int direct_native, int need_set_rs, 
 				  int multi_ok, int result_ignored, int nontail_self, int pop_and_jump, 
                                   int is_inlined, int unboxed_args);
@@ -1397,11 +1477,12 @@ int scheme_generate_finish_apply(mz_jit_state *jitter);
 int scheme_generate_finish_multi_apply(mz_jit_state *jitter);
 int scheme_generate_finish_tail_apply(mz_jit_state *jitter);
 void scheme_jit_register_sub_func(mz_jit_state *jitter, void *code, Scheme_Object *protocol);
-void scheme_jit_register_helper_func(mz_jit_state *jitter, void *code);
+void scheme_jit_register_helper_func(mz_jit_state *jitter, void *code, int gcable);
 #ifdef MZ_USE_FUTURES
 Scheme_Object *scheme_noncm_prim_indirect(Scheme_Prim proc, int argc);
 Scheme_Object *scheme_prim_indirect(Scheme_Primitive_Closure_Proc proc, int argc, Scheme_Object *self);
 #endif
+int scheme_generate_force_value_same_mark(mz_jit_state *jitter);
 
 /**********************************************************************/
 /*                             jitstack                               */
@@ -1443,6 +1524,8 @@ int scheme_generate(Scheme_Object *obj, mz_jit_state *jitter, int tail_ok, int w
                     Branch_Info *for_branch);
 int scheme_generate_unboxed(Scheme_Object *obj, mz_jit_state *jitter, int inlined_ok, int unbox_anyway);
 
+void scheme_generate_function_prolog(mz_jit_state *jitter);
+
 #ifdef USE_FLONUM_UNBOXING
 int scheme_generate_flonum_local_unboxing(mz_jit_state *jitter, int push, int no_store, int extfl);
 int scheme_generate_flonum_local_boxing(mz_jit_state *jitter, int pos, int offset, int target, int extfl);
@@ -1481,7 +1564,7 @@ int scheme_is_non_gc(Scheme_Object *obj, int depth);
 int scheme_jit_check_closure_flonum_bit(Scheme_Closure_Data *data, int pos, int delta);
 # define CLOSURE_ARGUMENT_IS_FLONUM(data, pos) scheme_jit_check_closure_flonum_bit(data, pos, 0)
 # define CLOSURE_CONTENT_IS_FLONUM(data, pos) scheme_jit_check_closure_flonum_bit(data, pos, data->num_params)
- int scheme_jit_check_closure_extflonum_bit(Scheme_Closure_Data *data, int pos, int delta);
+int scheme_jit_check_closure_extflonum_bit(Scheme_Closure_Data *data, int pos, int delta);
 # define CLOSURE_ARGUMENT_IS_EXTFLONUM(data, pos) scheme_jit_check_closure_extflonum_bit(data, pos, 0)
 # define CLOSURE_CONTENT_IS_EXTFLONUM(data, pos) scheme_jit_check_closure_extflonum_bit(data, pos, data->num_params)
 #endif
@@ -1571,7 +1654,3 @@ Scheme_Object *scheme_jit_continuation_apply_install(Apply_LWC_Args *args);
 #define INLINE_STRUCT_PROC_PROP_GET_W_DEFAULT 5
 #define INLINE_STRUCT_PROC_PROP_PRED 6
 #define INLINE_STRUCT_PROC_CONSTR 7
-
-
-
-
