@@ -1,6 +1,6 @@
 /*
   Racket
-  Copyright (c) 2004-2013 PLT Design Inc.
+  Copyright (c) 2004-2014 PLT Design Inc.
   Copyright (c) 1995-2001 Matthew Flatt
 
     This library is free software; you can redistribute it and/or
@@ -1903,6 +1903,14 @@ static void cons_onto_list(void *p)
 }
 #endif
 
+#if MZ_PRECISE_GC
+static int found_counter = 0;
+static void increment_found_counter(void *p)
+{
+  found_counter++;
+}
+#endif
+
 #if MZ_PRECISE_GC_TRACE
 static void count_struct_instance(void *p) {
   Scheme_Structure *s = (Scheme_Structure *)p;
@@ -2155,18 +2163,23 @@ Scheme_Object *scheme_dump_gc_stats(int c, Scheme_Object *p[])
   void *initial_trace_root = NULL;
   int (*inital_root_skip)(void *, size_t) = NULL;
 #endif
-#if MZ_PRECISE_GC_TRACE
+#ifdef MZ_PRECISE_GC
+  int skip_summary = 0;
   int trace_for_tag = 0;
-  int flags = 0;
-  int path_length_limit = 10000;
+  int dump_flags = 0;
   GC_for_each_found_proc for_each_found = NULL;
+# else
+#  define skip_summary 0
+#  define dump_flags 0
+#  define trace_for_tag 0
+#  define for_each_found NULL
+#endif
+#if MZ_PRECISE_GC_TRACE
+  int path_length_limit = 10000;
   GC_for_each_struct_proc for_each_struct = NULL;
 #else
 # ifndef USE_TAGGED_ALLOCATION
-#  define flags 0
-#  define trace_for_tag 0
 #  define path_length_limit 10000
-#  define for_each_found NULL
 #  define for_each_struct NULL
 #  define print_tagged_value NULL
 # endif
@@ -2431,6 +2444,29 @@ Scheme_Object *scheme_dump_gc_stats(int c, Scheme_Object *p[])
 
 #else
 
+# ifdef MZ_PRECISE_GC
+  if (c && SCHEME_SYMBOLP(p[0])) {
+    if (!strcmp("count", SCHEME_SYM_VAL(p[0]))
+        && (c == 2)
+        && SCHEME_SYMBOLP(p[1])) {
+      int i, maxpos;
+      maxpos = scheme_num_types();
+      for (i = 0; i < maxpos; i++) {
+        void *tn;
+        tn = scheme_get_type_name_or_null(i);
+        if (tn && !strcmp(tn, SCHEME_SYM_VAL(p[1]))) {
+          found_counter = 0;
+          for_each_found = increment_found_counter;
+          trace_for_tag = i;
+          dump_flags |= GC_DUMP_SUPPRESS_SUMMARY;
+          skip_summary = 1;
+          break;
+        }
+      }
+    }
+  }
+# endif
+
 # if MZ_PRECISE_GC_TRACE
   cons_accum_result = scheme_void;
   if (c && SCHEME_SYMBOLP(p[0])) {
@@ -2448,13 +2484,13 @@ Scheme_Object *scheme_dump_gc_stats(int c, Scheme_Object *p[])
       tn = scheme_get_type_name_or_null(i);
       if (tn && !strcmp(tn, s)) {
 	trace_for_tag = i;
-	flags |= GC_DUMP_SHOW_TRACE;
+	dump_flags |= GC_DUMP_SHOW_TRACE;
 	break;
       }
     }
 
     if (!strcmp("fnl", s))
-      flags |= GC_DUMP_SHOW_FINALS;
+      dump_flags |= GC_DUMP_SHOW_FINALS;
 
     if (!strcmp("struct", s)) {
       for_each_struct = count_struct_instance;
@@ -2493,7 +2529,7 @@ Scheme_Object *scheme_dump_gc_stats(int c, Scheme_Object *p[])
     }
   } else if (c && SCHEME_INTP(p[0])) {
     trace_for_tag = SCHEME_INT_VAL(p[0]);
-    flags |= GC_DUMP_SHOW_TRACE;
+    dump_flags |= GC_DUMP_SHOW_TRACE;
   } else if (c && SCHEME_THREADP(p[0])) {
     Scheme_Thread *t = (Scheme_Thread *)p[0];
     void **var_stack, *limit;
@@ -2528,15 +2564,15 @@ Scheme_Object *scheme_dump_gc_stats(int c, Scheme_Object *p[])
   else if ((c > 1) && SCHEME_SYMBOLP(p[1]) && !strcmp("cons", SCHEME_SYM_VAL(p[1]))) {
     for_each_found = cons_onto_list;
     cons_accum_result = scheme_null;
-    flags -= (flags & GC_DUMP_SHOW_TRACE);
+    dump_flags -= (dump_flags & GC_DUMP_SHOW_TRACE);
   }
-  scheme_console_printf("Begin Dump\n");
-#else
-  scheme_console_printf("Begin Dump\n");
 #endif
 
+  if (!skip_summary)
+    scheme_console_printf("Begin Dump\n");
+
 # ifdef MZ_PRECISE_GC
-  GC_dump_with_traces(flags, 
+  GC_dump_with_traces(dump_flags,
 		      scheme_get_type_name_or_null,
 		      for_each_found,
 		      trace_for_tag, trace_for_tag,
@@ -2623,27 +2659,40 @@ Scheme_Object *scheme_dump_gc_stats(int c, Scheme_Object *p[])
   }
 #endif
 
+  if (!skip_summary) {
+#ifdef MZ_PRECISE_GC
+    scheme_console_printf("Begin Help\n");
+    scheme_console_printf(" (dump-memory-stats 'count sym) - return number of instances of type named by sym.\n");
+    scheme_console_printf("   Example: (dump-memory-stats 'count '<pair>)\n");
 # if MZ_PRECISE_GC_TRACE
-  scheme_console_printf("Begin Help\n");
-  scheme_console_printf(" (dump-memory-stats sym) - prints paths to instances of type named by sym.\n");
-  scheme_console_printf("   Example: (dump-memory-stats '<pair>)\n");
-  scheme_console_printf(" (dump-memory-stats 'struct) - show counts for specific structure types.\n");
-  scheme_console_printf(" (dump-memory-stats 'fnl) - prints not-yet-finalized objects.\n");
-  scheme_console_printf(" (dump-memory-stats num) - prints paths to objects with tag num.\n");
-  scheme_console_printf(" (dump-memory-stats -num) - prints paths to objects of size num.\n");
-  scheme_console_printf(" (dump-memory-stats sym/num len) - limits path to size len.\n");
-  scheme_console_printf(" (dump-memory-stats sym/num 'cons) - builds list instead of showing paths.\n");
-  scheme_console_printf(" (dump-memory-stats 'peek num v) - returns value if num is address of object, v otherwise.\n");
-  scheme_console_printf(" (dump-memory-stats 'next v) - next tagged object after v, #f if none; start with #f.\n");
-  scheme_console_printf(" (dump-memory-stats 'addr v) - returns the address of v.\n");
-  scheme_console_printf(" (dump-memory-stats thread) - shows information about the thread.\n");
-  scheme_console_printf("End Help\n");
+    scheme_console_printf(" (dump-memory-stats sym) - prints paths to instances of type named by sym.\n");
+    scheme_console_printf("   Example: (dump-memory-stats '<pair>)\n");
+    scheme_console_printf(" (dump-memory-stats 'struct) - show counts for specific structure types.\n");
+    scheme_console_printf(" (dump-memory-stats 'fnl) - prints not-yet-finalized objects.\n");
+    scheme_console_printf(" (dump-memory-stats num) - prints paths to objects with tag num.\n");
+    scheme_console_printf(" (dump-memory-stats -num) - prints paths to objects of size num.\n");
+    scheme_console_printf(" (dump-memory-stats sym/num len) - limits path to size len.\n");
+    scheme_console_printf(" (dump-memory-stats sym/num 'cons) - builds list instead of showing paths.\n");
+    scheme_console_printf(" (dump-memory-stats 'peek num v) - returns value if num is address of object, v otherwise.\n");
+    scheme_console_printf(" (dump-memory-stats 'next v) - next tagged object after v, #f if none; start with #f.\n");
+    scheme_console_printf(" (dump-memory-stats 'addr v) - returns the address of v.\n");
+    scheme_console_printf(" (dump-memory-stats thread) - shows information about the thread.\n");
+# endif
+    scheme_console_printf("End Help\n");
+#endif
 
-  result = cons_accum_result;
+    scheme_console_printf("End Dump\n");
+  }
+
+#ifdef MZ_PRECISE_GC
+  if (for_each_found == increment_found_counter)
+    result = scheme_make_integer(found_counter);
+# if MZ_PRECISE_GC_TRACE
+  else
+    result = cons_accum_result;
   cons_accum_result = scheme_void;
 # endif
-
-  scheme_console_printf("End Dump\n");
+#endif
 
   scheme_end_atomic();
 
@@ -2911,7 +2960,7 @@ intptr_t scheme_count_memory(Scheme_Object *root, Scheme_Hash_Table *ht)
     break;
   case scheme_prim_type:
     {
-      if (((Scheme_Primitive_Proc *)root)->pp.flags & SCHEME_PRIM_IS_MULTI_RESULT)
+      if (SCHEME_PRIM_PROC_FLAGS(root) & SCHEME_PRIM_IS_MULTI_RESULT)
 	s = sizeof(Scheme_Prim_W_Result_Arity);
       else
 	s = sizeof(Scheme_Primitive_Proc);
@@ -2933,7 +2982,7 @@ intptr_t scheme_count_memory(Scheme_Object *root, Scheme_Hash_Table *ht)
     break;
   case scheme_closed_prim_type:
     {
-      if (((Scheme_Closed_Primitive_Proc *)root)->pp.flags & SCHEME_PRIM_IS_MULTI_RESULT)
+      if (SCHEME_PRIM_PROC_FLAGS(root) & SCHEME_PRIM_IS_MULTI_RESULT)
 	s = sizeof(Scheme_Closed_Prim_W_Result_Arity);
       else
 	s = sizeof(Scheme_Closed_Primitive_Proc);
