@@ -1,6 +1,6 @@
 /*
   Racket
-  Copyright (c) 2004-2013 PLT Design Inc.
+  Copyright (c) 2004-2014 PLT Design Inc.
   Copyright (c) 1995-2001 Matthew Flatt
 
     This library is free software; you can redistribute it and/or
@@ -39,6 +39,8 @@ static Scheme_Object *make_channel(int n, Scheme_Object **p);
 static Scheme_Object *make_channel_put(int n, Scheme_Object **p);
 static Scheme_Object *channel_p(int n, Scheme_Object **p);
 static Scheme_Object *channel_put_p(int n, Scheme_Object **p);
+static Scheme_Object *chaperone_channel(int argc, Scheme_Object *argv[]);
+static Scheme_Object *impersonate_channel(int argc, Scheme_Object *argv[]);
 
 static Scheme_Object *thread_send(int n, Scheme_Object **p);
 static Scheme_Object *thread_receive(int n, Scheme_Object **p);
@@ -154,6 +156,16 @@ void scheme_init_sema(Scheme_Env *env)
                                                       "channel-put-evt?",
                                                       1, 1, 1), 
 			     env);
+  scheme_add_global_constant("chaperone-channel",
+			     scheme_make_prim_w_arity(chaperone_channel,
+                                                      "chaperone-channel",
+                                                      3, -1),
+			     env);
+  scheme_add_global_constant("impersonate-channel",
+			     scheme_make_prim_w_arity(impersonate_channel,
+                                                      "impersonate-channel",
+                                                      3, -1),
+			     env);
 
   scheme_add_global_constant("thread-send", 
 			     scheme_make_prim_w_arity(thread_send,
@@ -1042,17 +1054,62 @@ int scheme_try_channel_put(Scheme_Object *ch, Scheme_Object *v)
     return 0;
 }
 
+Scheme_Object *chaperone_put(Scheme_Object *obj, Scheme_Object *orig)
+{
+  Scheme_Chaperone *px = (Scheme_Chaperone *)obj;
+  Scheme_Object *val = orig;
+  Scheme_Object *a[2];
+  Scheme_Object *redirect;
+
+  while (1) {
+    if (SCHEME_CHANNELP(px)) {
+      return val;
+    } else if (!(SAME_TYPE(SCHEME_TYPE(px->redirects), scheme_nack_guard_evt_type))) {
+      a[0] = px->prev;
+      a[1] = val;
+      redirect = px->redirects;
+      val = _scheme_apply(redirect, 2, a);
+
+      if (!(SCHEME_CHAPERONE_FLAGS(px) & SCHEME_CHAPERONE_IS_IMPERSONATOR))
+        if (!scheme_chaperone_of(val, orig))
+          scheme_wrong_chaperoned("channel-put", "result", orig, val);
+
+      px = (Scheme_Chaperone *)px->prev;
+    } else {
+      /* In this case, the `px` is actually an evt chaperone so we
+         don't want to handle it here since we're doing a "put" */
+      px = (Scheme_Chaperone *)px->prev;
+    }
+  }
+
+  return obj;
+}
+
 static Scheme_Object *make_channel_put(int argc, Scheme_Object **argv)
 {
-  if (!SCHEME_CHANNELP(argv[0]))
-    scheme_wrong_contract("channel-put-evt", "channel?", 0, argc, argv);
+  Scheme_Object *ch = argv[0];
+  Scheme_Object *val = argv[1];
+  Scheme_Object *chaperone = NULL;
 
-  return scheme_make_channel_put_evt(argv[0], argv[1]);
+  if (SCHEME_NP_CHAPERONEP(ch)
+      && SCHEME_CHANNELP(SCHEME_CHAPERONE_VAL(ch))) {
+    chaperone = ch;
+    ch = SCHEME_CHAPERONE_VAL(chaperone);
+  } else if (!SCHEME_CHANNELP(argv[0])) {
+    scheme_wrong_contract("channel-put-evt", "channel?", 0, argc, argv);
+  }
+
+  if (chaperone)
+    val = chaperone_put(chaperone, argv[1]);
+
+  return scheme_make_channel_put_evt(ch, val);
 }
 
 static Scheme_Object *channel_p(int n, Scheme_Object **p)
 {
-  return (SCHEME_CHANNELP(p[0])
+  return ((SCHEME_CHANNELP(p[0]) ||
+           (SCHEME_NP_CHAPERONEP(p[0])
+            && SCHEME_CHANNELP(SCHEME_CHAPERONE_VAL(p[0]))))
 	  ? scheme_true
 	  : scheme_false);
 }
@@ -1109,6 +1166,51 @@ int scheme_try_channel_get(Scheme_Object *ch)
   return 0;
 }
 
+/* This chaperone only protects the "put" end of the channel because
+   chaperone-evt is sufficient to protect the "get" end. Thus, it first
+   wraps the object in an evt chaperone. */
+Scheme_Object *do_chaperone_channel(const char *name, int is_impersonator, int argc, Scheme_Object **argv)
+{
+  Scheme_Chaperone *px;
+  Scheme_Object *val = argv[0];
+  Scheme_Object *evt;
+  Scheme_Hash_Tree *props;
+
+  if (SCHEME_CHAPERONEP(val))
+    val = SCHEME_CHAPERONE_VAL(val);
+
+  if (!SCHEME_CHANNELP(val))
+    scheme_wrong_contract(name, "channel?", 0, argc, argv);
+  scheme_check_proc_arity(name, 1, 1, argc, argv);
+  scheme_check_proc_arity(name, 2, 2, argc, argv);
+
+  evt = scheme_do_chaperone_evt(name, is_impersonator, 2, argv);
+
+  props = scheme_parse_chaperone_props(name, 3, argc, argv);
+
+  px = MALLOC_ONE_TAGGED(Scheme_Chaperone);
+  px->iso.so.type = scheme_chaperone_type;
+  px->val = val;
+  px->prev = evt;
+  px->props = props;
+  px->redirects = argv[2];
+
+  if (is_impersonator)
+    SCHEME_CHAPERONE_FLAGS(px) |= SCHEME_CHAPERONE_IS_IMPERSONATOR;
+
+  return (Scheme_Object *)px;
+}
+
+static Scheme_Object *chaperone_channel(int argc, Scheme_Object **argv)
+{
+  return do_chaperone_channel("chaperone-channel", 0, argc, argv);
+}
+
+static Scheme_Object *impersonate_channel(int argc, Scheme_Object **argv)
+{
+  return do_chaperone_channel("impersonator-channel", 1, argc, argv);
+}
+
 /**********************************************************************/
 /*                           Thread mbox                              */
 /**********************************************************************/
