@@ -199,92 +199,100 @@ enum {
   scheme_place_bi_channel_type,         /* 175 */
   scheme_once_used_type,                /* 176 */
   scheme_serialized_symbol_type,        /* 177 */
-  scheme_serialized_structure_type,     /* 178 */
-  scheme_fsemaphore_type,               /* 179 */
-  scheme_serialized_tcp_fd_type,        /* 180 */
-  scheme_serialized_file_fd_type,       /* 181 */
-  scheme_port_closed_evt_type,          /* 182 */
-  scheme_proc_shape_type,               /* 183 */
-  scheme_struct_proc_shape_type,        /* 184 */
-  scheme_phantom_bytes_type,            /* 185 */
+  scheme_serialized_keyword_type,       /* 178 */
+  scheme_serialized_structure_type,     /* 179 */
+  scheme_fsemaphore_type,               /* 180 */
+  scheme_serialized_tcp_fd_type,        /* 181 */
+  scheme_serialized_file_fd_type,       /* 182 */
+  scheme_port_closed_evt_type,          /* 183 */
+  scheme_proc_shape_type,               /* 184 */
+  scheme_struct_proc_shape_type,        /* 185 */
+  scheme_phantom_bytes_type,            /* 186 */
+  scheme_environment_variables_type,    /* 187 */
+  scheme_filesystem_change_evt_type,    /* 188 */
+  scheme_ctype_type,                    /* 189 */
+  scheme_plumber_type,                  /* 190 */
+  scheme_plumber_handle_type,           /* 191 */
                                         
 #ifdef MZTAG_REQUIRED                            
-  _scheme_last_normal_type_,            /* 186 */
+  _scheme_last_normal_type_,            /* 192 */
                                                  
-  scheme_rt_weak_array,                 /* 187 */
+  scheme_rt_weak_array,                 /* 193 */
                                         
-  scheme_rt_comp_env,                   /* 188 */
-  scheme_rt_constant_binding,           /* 189 */
-  scheme_rt_resolve_info,               /* 190 */
-  scheme_rt_unresolve_info,             /* 191 */
-  scheme_rt_optimize_info,              /* 192 */
-  scheme_rt_compile_info,               /* 193 */
-  scheme_rt_cont_mark,                  /* 194 */
-  scheme_rt_saved_stack,                /* 195 */
-  scheme_rt_reply_item,                 /* 196 */
-  scheme_rt_closure_info,               /* 197 */
-  scheme_rt_overflow,                   /* 198 */
-  scheme_rt_overflow_jmp,               /* 199 */
-  scheme_rt_meta_cont,                  /* 200 */
-  scheme_rt_dyn_wind_cell,              /* 201 */
-  scheme_rt_dyn_wind_info,              /* 202 */
-  scheme_rt_dyn_wind,                   /* 203 */
-  scheme_rt_dup_check,                  /* 204 */
-  scheme_rt_thread_memory,              /* 205 */
-  scheme_rt_input_file,                 /* 206 */
-  scheme_rt_input_fd,                   /* 207 */
-  scheme_rt_oskit_console_input,        /* 208 */
-  scheme_rt_tested_input_file,          /* 209 */
-  scheme_rt_tested_output_file,         /* 210 */
-  scheme_rt_indexed_string,             /* 211 */
-  scheme_rt_output_file,                /* 212 */
-  scheme_rt_load_handler_data,          /* 213 */
-  scheme_rt_pipe,                       /* 214 */
-  scheme_rt_beos_process,               /* 215 */
-  scheme_rt_system_child,               /* 216 */
-  scheme_rt_tcp,                        /* 217 */
-  scheme_rt_write_data,                 /* 218 */
-  scheme_rt_tcp_select_info,            /* 219 */
-  scheme_rt_param_data,                 /* 220 */
-  scheme_rt_will,                       /* 221 */
-  scheme_rt_linker_name,                /* 222 */
-  scheme_rt_param_map,                  /* 223 */
-  scheme_rt_finalization,               /* 224 */
-  scheme_rt_finalizations,              /* 225 */
-  scheme_rt_cpp_object,                 /* 226 */
-  scheme_rt_cpp_array_object,           /* 227 */
-  scheme_rt_stack_object,               /* 228 */
-  scheme_rt_preallocated_object,        /* 229 */
-  scheme_thread_hop_type,               /* 230 */
-  scheme_rt_srcloc,                     /* 231 */
-  scheme_rt_evt,                        /* 232 */
-  scheme_rt_syncing,                    /* 233 */
-  scheme_rt_comp_prefix,                /* 234 */
-  scheme_rt_user_input,                 /* 235 */
-  scheme_rt_user_output,                /* 236 */
-  scheme_rt_compact_port,               /* 237 */
-  scheme_rt_read_special_dw,            /* 238 */
-  scheme_rt_regwork,                    /* 239 */
-  scheme_rt_rx_lazy_string,             /* 240 */
-  scheme_rt_buf_holder,                 /* 241 */
-  scheme_rt_parameterization,           /* 242 */
-  scheme_rt_print_params,               /* 243 */
-  scheme_rt_read_params,                /* 244 */
-  scheme_rt_native_code,                /* 245 */
-  scheme_rt_native_code_plus_case,      /* 246 */
-  scheme_rt_jitter_data,                /* 247 */
-  scheme_rt_module_exports,             /* 248 */
-  scheme_rt_delay_load_info,            /* 249 */
-  scheme_rt_marshal_info,               /* 250 */
-  scheme_rt_unmarshal_info,             /* 251 */
-  scheme_rt_runstack,                   /* 252 */
-  scheme_rt_sfs_info,                   /* 253 */
-  scheme_rt_validate_clearing,          /* 254 */
-  scheme_rt_avl_node,                   /* 255 */
-  scheme_rt_lightweight_cont,           /* 256 */
-  scheme_rt_export_info,                /* 257 */
-  scheme_rt_cont_jmp,                   /* 258 */
+  scheme_rt_comp_env,                   /* 194 */
+  scheme_rt_constant_binding,           /* 195 */
+  scheme_rt_resolve_info,               /* 196 */
+  scheme_rt_unresolve_info,             /* 197 */
+  scheme_rt_optimize_info,              /* 198 */
+  scheme_rt_compile_info,               /* 199 */
+  scheme_rt_cont_mark,                  /* 200 */
+  scheme_rt_saved_stack,                /* 201 */
+  scheme_rt_reply_item,                 /* 202 */
+  scheme_rt_closure_info,               /* 203 */
+  scheme_rt_overflow,                   /* 204 */
+  scheme_rt_overflow_jmp,               /* 205 */
+  scheme_rt_meta_cont,                  /* 206 */
+  scheme_rt_dyn_wind_cell,              /* 207 */
+  scheme_rt_dyn_wind_info,              /* 208 */
+  scheme_rt_dyn_wind,                   /* 209 */
+  scheme_rt_dup_check,                  /* 210 */
+  scheme_rt_thread_memory,              /* 211 */
+  scheme_rt_input_file,                 /* 212 */
+  scheme_rt_input_fd,                   /* 213 */
+  scheme_rt_oskit_console_input,        /* 214 */
+  scheme_rt_tested_input_file,          /* 215 */
+  scheme_rt_tested_output_file,         /* 216 */
+  scheme_rt_indexed_string,             /* 217 */
+  scheme_rt_output_file,                /* 218 */
+  scheme_rt_load_handler_data,          /* 219 */
+  scheme_rt_pipe,                       /* 220 */
+  scheme_rt_beos_process,               /* 221 */
+  scheme_rt_system_child,               /* 222 */
+  scheme_rt_tcp,                        /* 223 */
+  scheme_rt_write_data,                 /* 224 */
+  scheme_rt_tcp_select_info,            /* 225 */
+  scheme_rt_param_data,                 /* 226 */
+  scheme_rt_will,                       /* 227 */
+  scheme_rt_linker_name,                /* 228 */
+  scheme_rt_param_map,                  /* 229 */
+  scheme_rt_finalization,               /* 230 */
+  scheme_rt_finalizations,              /* 231 */
+  scheme_rt_cpp_object,                 /* 232 */
+  scheme_rt_cpp_array_object,           /* 233 */
+  scheme_rt_stack_object,               /* 234 */
+  scheme_rt_preallocated_object,        /* 235 */
+  scheme_thread_hop_type,               /* 236 */
+  scheme_rt_srcloc,                     /* 237 */
+  scheme_rt_evt,                        /* 238 */
+  scheme_rt_syncing,                    /* 239 */
+  scheme_rt_comp_prefix,                /* 240 */
+  scheme_rt_user_input,                 /* 241 */
+  scheme_rt_user_output,                /* 242 */
+  scheme_rt_compact_port,               /* 243 */
+  scheme_rt_read_special_dw,            /* 244 */
+  scheme_rt_regwork,                    /* 245 */
+  scheme_rt_rx_lazy_string,             /* 246 */
+  scheme_rt_buf_holder,                 /* 247 */
+  scheme_rt_parameterization,           /* 248 */
+  scheme_rt_print_params,               /* 249 */
+  scheme_rt_read_params,                /* 250 */
+  scheme_rt_native_code,                /* 251 */
+  scheme_rt_native_code_plus_case,      /* 252 */
+  scheme_rt_jitter_data,                /* 253 */
+  scheme_rt_module_exports,             /* 254 */
+  scheme_rt_delay_load_info,            /* 255 */
+  scheme_rt_marshal_info,               /* 256 */
+  scheme_rt_unmarshal_info,             /* 257 */
+  scheme_rt_runstack,                   /* 258 */
+  scheme_rt_sfs_info,                   /* 259 */
+  scheme_rt_validate_clearing,          /* 260 */
+  scheme_rt_avl_node,                   /* 261 */
+  scheme_rt_lightweight_cont,           /* 262 */
+  scheme_rt_export_info,                /* 263 */
+  scheme_rt_cont_jmp,                   /* 264 */
+  scheme_rt_letrec_check_frame,         /* 265 */
 #endif
+  scheme_deferred_expr_type,            /* 266 */
 
   _scheme_last_type_
 };
